/*
 * Written by Artur Biesiadowski <abies@pg.gda.pl>
 * This file is public domain - you can use/modify/distribute it as long
 * as some credit is given to me. You are not required to keep it open
 * sourced, nor to give back all changes to me, BUT if you do, everybody
 * will benefit.
 * For latest version contact me at <abies@pg.gda.pl> or check
 * http://www.gjt.org/servlets/JCVSlet/list/gjt/top.org.gjt.abies
 * This file comes with no guarantee of anything - you have been WARNED.
 */

package org.gjt.abies;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.NoSuchElementException;

public class Queue implements Serializable, Cloneable
{
	private transient Object[] data;
	private transient int offset;
	private transient int count;

	public Queue()
	{
		this(10);
	}

	public Queue(int size)
	{
		if ( size <= 0 )
			throw new IllegalArgumentException("Size has to be greater than 0");
		data = new Object[size];
	}

	public int size()
	{
		return count;
	}

	public Object peekHead()
	{
		if ( count == 0 )
			throw new NoSuchElementException();
		return data[offset];
	}

	public Object peekTail()
	{
		if ( count == 0 )
			throw new NoSuchElementException();
		return data[correctIndex(count-1)];
	}

	public Object removeHead()
	{
		if ( count == 0 )
			throw new NoSuchElementException();
		Object r = data[offset];
		data[offset] = null;
		offset++;
		count--;
		if ( offset >= data.length )
		{
			if ( offset > data.length)
				throw new InternalError("Invariant violated at Queue - offset greater than array length");
			offset = 0;
		}
		return r;
	}

	public Object removeTail()
	{
		if ( count == 0 )
			throw new NoSuchElementException();
		int index = correctIndex(count-1);
		Object r = data[index];
		data[index] = null;
		count--;
		return r;
	}

	public void addHead(Object obj)
	{
		if ( count == data.length )
			setSize(count*2);
		if ( offset == 0 )
		{
			offset = data.length-1;
		}
		else
		{
			offset--;
		}
		data[offset] = obj;
		count++;
	}

	public void addTail(Object obj)
	{
		if ( count == data.length )
			setSize(count*2);
		int index = correctIndex(count);
		data[index] = obj;
		count++;
	}

	public boolean isEmpty()
	{
		return count == 0;
	}

	public void trim()
	{
		if ( count == data.length )
			return;
		setSize(count);
	}

	private int correctIndex(int index)
	{
		return (offset+index) % data.length;
	}

	private void setSize(int nsize)
	{
		if ( nsize < count )
			throw new InternalError("Precondition violated: setSize called with size smaller than count");
		Object[] ndata = new Object[nsize];
		int tail = Math.min(count,data.length-offset);
		System.arraycopy(data,offset,ndata,0,tail);
		System.arraycopy(data,0,ndata,tail,count -tail);
		data = ndata;
		offset = 0;
	}

	private void writeObject(ObjectOutputStream oOut)
		throws IOException
	{
		oOut.defaultWriteObject();
		oOut.writeInt(count);
		oOut.writeInt(data.length);
		for ( int i =0; i < count; i++ )
		{
			int index = correctIndex(i);
			oOut.writeObject(data[index]);
		}
	}

	private void readObject(ObjectInputStream oIn)
		throws IOException, ClassNotFoundException
	{
		oIn.defaultReadObject();
		count = oIn.readInt();
		data = new Object[oIn.readInt()];
		offset = 0;
		for ( int i =0; i < count; i++ )
		{
			data[i] = oIn.readObject();
		}
	}


	public boolean equals(Object obj)
	{
		return obj instanceof Queue && equals((Queue)obj);
	}

	public boolean equals(Queue q)
	{
		if ( q.count != count )
			return false;
		for ( int i =0; i < count; i++ )
		{
			Object o1 = data[correctIndex(i)];
			Object o2 = q.data[q.correctIndex(i)];
			if ( notEqual(o1,o2) )
				return false;
		}
		return true;
	}

	public String toString()
	{
		StringBuffer sb = new StringBuffer();
		sb.append('[');
		for ( int i =0; i < count; i++)
		{
			sb.append(data[correctIndex(i)]);
			if ( i < count-1)
				sb.append(',');
		}
		sb.append(']');
		return sb.toString();
	}

	public Object clone() throws CloneNotSupportedException
	{
		Queue q = (Queue)super.clone();
		q.data = (Object[])q.data.clone();
		return q;
	}



	private static boolean notEqual(Object a, Object b)
	{
		return ((a == null) ? (b != null) : !a.equals(b));
	}


	private static final long serialVersionUID = 0;



	// TESTING

	public static void main(String argv[]) throws CloneNotSupportedException
	{
		Queue q = new Queue(1);
		System.out.println("Empty: " + q);
		System.out.println("TRUE:" + q.isEmpty());
		q.addHead("5");
		q.addHead("4");
		q.addTail("6");
		q.addTail("7");
		q.addHead("3");
		q.addTail("8");
		q.addTail("9");
		q.addTail("10");
		System.out.println(q);
		Queue q2 = (Queue)q.clone();
		while ( !q2.isEmpty() )
		{
			System.out.print(q2.removeHead()+" ");
		}
		System.out.println();
		q2 = (Queue)q.clone();
		while ( !q2.isEmpty() )
		{
			System.out.print(q2.removeTail()+" ");
		}
		System.out.println();
		try {
			q2.removeTail();
			throw new RuntimeException("NoSuchElement not thrown");
		} catch (NoSuchElementException exc)
			{
			}
		try {
			q2.removeHead();
			throw new RuntimeException("NoSuchElement not thrown");
		} catch (NoSuchElementException exc)
			{
			}
		q2 = (Queue)q.clone();
		System.out.println("TRUE:"+q2.equals(q));
		q2.addHead("Something");
		System.out.println("FALSE:"+ q2.equals(q));
	}



}